/****************************************************************************\
*   BIEWLIB - this header file is part of Binary vIEW project (BIEW)         *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   biewlib/biewlib.h                                  *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain extensions of standard C library, that needed for BIEW  *
* project.                                                                   *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#ifndef __BIEWLIB_H
#define __BIEWLIB_H 1

typedef enum { False = 0, True = 1 }tBool;

#ifndef __NORECURSIVE
#ifndef __SYS_DEP_H
#include "biewlib/sysdep/_sys_dep.h"
#endif

#ifndef __OS_DEP_H
#include "biewlib/sysdep/__os_dep.h"
#endif

#ifndef __HRD_INF_H
#include "biewlib/sysdep/_hrd_inf.h"
#endif

#endif

#ifdef __cplusplus
extern "C" {
#endif

/* Using own code for qsort and bsearch functions is guarantee of stable work */

typedef int tCompare;
typedef tCompare (__FASTCALL__ *func_compare)(const void __HUGE__ *e1,const void __HUGE__ *e2);

extern void  __FASTCALL__ HQSort(void __HUGE__ *base, unsigned long num, unsigned width,
                                 func_compare);
extern void __HUGE__ * __FASTCALL__ HLFind(const void *key,
                                     void __HUGE__ *base,
                                     unsigned long nelem,unsigned width,
                                     func_compare);

/* return nearest element of array toward zero */
extern unsigned long __FASTCALL__ HLFindNearest(const void *key,
                                     void __HUGE__ *base,
                                     unsigned long nelem,unsigned width,
                                     func_compare);

extern tBool  __FASTCALL__ isseparate(int ch);

/* ASCIIZ string extended support */

extern int   __FASTCALL__ szTrimTrailingSpace(char *str);
extern int   __FASTCALL__ szTrimLeadingSpace(char *str);
extern void  __FASTCALL__ szSpace2Tab(char *dest,const char *src);
extern int   __FASTCALL__ szTab2Space(char *dest,const char *src);
extern char *__FASTCALL__ szKillSpaceAround(char *str,char *point_to);

/*
  Only this function must be used for error reporting
	(do not use printf, fprintf, etc. !)
*/
extern int  printm(const char *str,...);

/* Linear memory container */
typedef struct tag_linearArray
{
  unsigned long   nItems;
  void __HUGE__ * data;
  unsigned long   nSize;
  unsigned        itemSize;
}linearArray;

extern linearArray *__FASTCALL__ la_Build( unsigned long maxitems,unsigned size_of_item,
                                           void (__FASTCALL__ *mem_out)(const char *));
extern void __HUGE__*__FASTCALL__ la_AddData(linearArray *obj,const void *data,void (__FASTCALL__ *mem_out)(const char *));
extern void         __FASTCALL__ la_Destroy(linearArray *obj);
extern void         __FASTCALL__ la_IterDestroy(linearArray *obj,void (__FASTCALL__ *del_func)(void __HUGE__ *));
extern void         __FASTCALL__ la_ForEach(linearArray *obj,void (__FASTCALL__ *iter_func)(void __HUGE__ *));
extern void         __FASTCALL__ la_Sort(linearArray *obj,func_compare);
extern void __HUGE__ *__FASTCALL__ la_Find(linearArray *obj,const void *key,
                                           func_compare);
extern unsigned long __FASTCALL__ la_FindNearest(linearArray *obj, const void *key,
                                           func_compare);


#ifdef __cplusplus
}
#endif

#endif




