/****************************************************************************\
*   FILE_INI - this header file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   biewlib/file_ini.h                                 *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain prototypes of .ini file services.                       *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
/****************************************************************************\
*                         File_ini     Version 1.00                          *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
* File_ini.h - header file for software support work with files have         *
* extension [.ini]; use self syntax for it's files.                          *
*----------------------------------------------------------------------------*
* Preface : This format enabled full initialization of programms and programm*
*           groups. Easy to use and understand like a traditional MS Windows *
*           form, but have expanded syntax and simple C-like preprocessor.   *
*============================================================================*
* Helpful diagramm :                                                         *
*    ; - symbol of commentaries from here and to end of line.                *
*  [ xx ] - Section. Section is global concept, may contain items and        *
*           subsections                                                      *
*  < xx > - SubSection. SubSection is local concept, may contain items only. *
*  xx=yy  - Item. Item is named variable, that contain value that used in    *
*           program.                                                         *
* #xx    -  Preprocessor instructions. Internal command for [.ini] file      *
*           interpretator.                                                   *
\****************************************************************************/

#ifndef __FILE_INI_RUNTIME_SUPPORT_SYSTEM__
#define __FILE_INI_RUNTIME_SUPPORT_SYSTEM__

#ifndef __BBIO_H
#include "biewlib/bbio.h"
#endif

/*
    possible errors that are generic
*/

#define __FI_NOERRORS      0
#define __FI_BADFILENAME  -1
#define __FI_TOOMANY      -2
#define __FI_NOTMEM       -3
#define __FI_OPENCOND     -4
#define __FI_IFNOTFOUND   -5
#define __FI_ELSESTAT     -6
#define __FI_UNRECOGN     -7
#define __FI_BADCOND      -8
#define __FI_OPENSECT     -9
#define __FI_BADCHAR      -10
#define __FI_BADVAR       -11
#define __FI_BADVAL       -12
#define __FI_NOVAR        -13
#define __FI_NODEFVAR     -14
#define __FI_ELIFSTAT     -15
#define __FI_OPENVAR      -16
#define __FI_NOTEQU       -17
#define __FI_USER         -18
#define __FI_FIUSER       -19

/*
    possible answers to the errors
*/

#define __FI_IGNORE   0
#define __FI_EXITPROC 1

/*
    return constants for FiSearch
*/

#define __FI_NOTFOUND   0
#define __FI_SECTION    1
#define __FI_SUBSECTION 2
#define __FI_ITEM       3

#ifdef __cplusplus
extern "C" {
#endif

typedef BGLOBAL FiHandler;

typedef struct tagIniInfo
{
  const char * section;
  const char * subsection;
  const char * item;
  const char * value;
}IniInfo;

/* Must return False - for continue   True - all done */
typedef tBool      (__FASTCALL__ *FiUserFunc)(IniInfo * info);

/******************************************************\
* You can exchange all this pointers to self routines  *
*    and release virtual access to the methods         *
*                                                      *
* Chematic disgramm :                                  *
*                                                      *
*   FiFileProcessor   -- call --> FiStringProcessor    *
*   FiStringProcessor -- call --> FiCommandProcessor   *
*   FiCommandProcessor - call --> FiFileProcessor &    *
*                                 FiStringProcessor    *
\******************************************************/

extern  int    __FASTCALL__ (*FiError)(int nError,int row);
extern  void   __FASTCALL__ (*FiFileProcessor)(const char *fname);
extern  tBool  __FASTCALL__ (*FiStringProcessor)(char * curr_str);
extern  tBool  __FASTCALL__ (*FiCommandProcessor)(const char * cmd);
extern  tBool  __FASTCALL__ (*FiGetCondition)(const char * cond);

#define FI_MAXSTRLEN 255


extern const char *  __FASTCALL__ FiDecodeError(int nError);
extern unsigned int  *FinCurrString;
extern char **       FiFileNames;
extern unsigned int  FiFilePtr;

extern tBool         FiAllWantInput ; /* All input exclude commentaries, i.e. carriage return and line feed and space characters will be returned */

extern char *        FiUserMessage;

extern char FiOpenComment; /* comment always start with FiOpenComment symbol and to end of line */

void         __FASTCALL__ FiAError(int nError,int row);
void         __FASTCALL__ FiAErrorCL(int nError); /* error in curent line */

/********************************************************************\
* High Level procedure                                               *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
* Use only this function.                                            *
* This function calls UserFunc as CALLBACK routine.                  *
* All internal variables will be expanded, and command processor     *
* will be done. Unknown instruction and hotkeys should be ignored.   *
\********************************************************************/

void          __FASTCALL__ FiProgress(const char * filename,FiUserFunc);

/******************************************************************\
* Low level routines                                               *
\******************************************************************/

FiHandler     __FASTCALL__ FiOpen( const char * filename );
void          __FASTCALL__ FiClose( FiHandler h );
int           __FASTCALL__ FiSearch( FiHandler h, const char * Name );
void          __FASTCALL__ FiSeekTo( FiHandler h, int nSection, int nSubSection, int nItem );
char *        __FASTCALL__ FiGetNextString( FiHandler h, char * store, unsigned int len, char *original );
unsigned int  __FASTCALL__ FiGetNumberOfSections( FiHandler h);
unsigned int  __FASTCALL__ FiGetTotalNumberOfSubSections( FiHandler h );
unsigned int  __FASTCALL__ FiGetLocalNumberOfSubSections( FiHandler h, int nSection );
unsigned int  __FASTCALL__ FiGetTotalNumberOfItems( FiHandler h);
unsigned int  __FASTCALL__ FiGetLocalNumberOfItems( FiHandler h,int nSection , int nSubSection);

tBool         __FASTCALL__ FiisSection( const char * str );
tBool         __FASTCALL__ FiisSubSection( const char * str );
tBool         __FASTCALL__ FiisItem( const char * str);
tBool         __FASTCALL__ FiisCommand( const char * str);

unsigned int  __FASTCALL__ FiGetLengthSection( const char * src );
unsigned int  __FASTCALL__ FiGetLengthSubSection( const char * src );
unsigned int  __FASTCALL__ FiGetLengthItem( const char * src );
unsigned int  __FASTCALL__ FiGetLengthValue( const char * src );
unsigned int  __FASTCALL__ FiGetLengthCommandString( const char * src );

char *        __FASTCALL__ FiGetSectionName(const char * src,char * store);
char *        __FASTCALL__ FiGetSubSectionName(const char * src, char * store);
char *        __FASTCALL__ FiGetItemName(const char * src, char * store);
char *        __FASTCALL__ FiGetValueOfItem(const char * src, char * store);
char *        __FASTCALL__ FiGetCommandString(const char * src, char * store);

void          __FASTCALL__ FiFileProcessorStd( const char * filename );
tBool         __FASTCALL__ FiStringProcessorStd( char * string );
tBool         __FASTCALL__ FiCommandProcessorStd( const char * cmd );

/*
    WORD processor
*/

typedef struct tagSTRING
{
  const char * str;
  unsigned int iptr;
}STRING;

unsigned int  __FASTCALL__ FiGetLengthBracketString( const char * src );
char *        __FASTCALL__ FiGetBracketString(const char * str, char * store);
int           __FASTCALL__ FiisLegal(const char * illegal,char c);
unsigned int  __FASTCALL__ FiGetLengthNextWord( STRING * str, const char * illegal_symbols);
char *        __FASTCALL__ FiGetNextWord( STRING * str,const char * illegal_symbols,char * store );

/*
    variables set
*/

typedef struct tagVar
{
  char * variables;
  char * associate;
  struct tagVar * next;
  struct tagVar * prev;
}Var;

typedef Var * pVar;

pVar            __FASTCALL__ FiConstructVar(const char *v,const char *a);
void            __FASTCALL__ FiDeleteVar(pVar pp);
void            __FASTCALL__ FiDeleteAllVar( void );
const char    * __FASTCALL__ FiExpandVariables(const char * var);
tBool           __FASTCALL__ FiExpandAllVar(const char * value,char * store);
void            __FASTCALL__ FiAddVariables(const char * var,const char * associate);
void            __FASTCALL__ FiRemoveVariables(const char * var);

tBool           __FASTCALL__ FiGetConditionStd( const char *condstr);

/*
    High level routines (similar to MS WIN SDK)
*/

#define HINI_FULLCACHED 0x0001
#define HINI_UPDATED    0x0002

typedef struct tag_iniProfile
{
   FiHandler     handler;
   char *        fname;
   void *        cache;
   unsigned      flags;
}hIniProfile;

/* For internal purposes */
extern void __FASTCALL__ hlFiProgress(hIniProfile *ini,FiUserFunc usrproc);

/* For public use */

extern hIniProfile*    __FASTCALL__ iniOpenFile(const char *fname,tBool *has_error);
extern void            __FASTCALL__ iniCloseFile(hIniProfile *ini);

extern unsigned __FASTCALL__ iniReadProfileString(hIniProfile *ini,
                                     const char *section,
                                     const char *subsection,
                                     const char *_item,
                                     const char *def_value,
                                     char *buffer,
                                     unsigned cbBuffer);

extern tBool __FASTCALL__ iniWriteProfileString(hIniProfile *ini,
                                     const char *section,
                                     const char *subsection,
                                     const char *item,
                                     const char *value);

#ifdef __cplusplus
}
#endif

#endif




