/****************************************************************************\
*   DISASM - this source file is part of Binary vIEW project (BIEW)          *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/disasm.c                                   *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of Intel disassembler (main module).     *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <errno.h>

#include "colorset.h"
#include "bmfile.h"
#include "plugins/hexmode.h"
#include "plugins/disasm.h"
#include "biewutil.h"
#include "reg_form.h"
#include "bconsole.h"
#include "editor.h"
#include "codeguid.h"
#include "biewlib/file_ini.h"
#include "biewlib/biewlib.h"
#include "biewlib/pmalloc.h"
#include "biewlib/kbd_code.h"

/* references resolution */
#define NEEDREF_ALL 2
#define NEEDREF_JMP 1
#define NEEDREF_NONE 0

extern REGISTRY_DISASM ix86_Disasm;
extern REGISTRY_DISASM Null_Disasm;

static REGISTRY_DISASM *mainDisasmTable[] =
{
  &ix86_Disasm,
  &Null_Disasm
};

static unsigned DefDisasmSel = __DEFAULT_DISASM;
REGISTRY_DISASM *activeDisasm = NULL;

static unsigned long PrevPageSize,CurrPageSize,PrevStrLen,CurrStrLen;
static unsigned disNeedRef = 0;
int DisasmCurrLine;
unsigned disPanelMode = 0;
static int           HiLight = 1;
static char          *CurrStrLenBuff = NULL;
static unsigned long *PrevStrLenAddr = NULL;
static char          LastPrevLen;
static char          PrevStrCount = 0;
static char *        disCodeBuffer = NULL;
static int           disMaxCodeLen;

tBool DisasmPrepareMode = False;

static void __NEAR__ __FASTCALL__ disAcceptActions( void );

static tBool __FASTCALL__ disSelect_Disasm( void )
{
  const char *modeName[sizeof(mainDisasmTable)/sizeof(REGISTRY_DISASM *)];
  size_t i,nModes;
  int retval;

  nModes = sizeof(mainDisasmTable)/sizeof(REGISTRY_DISASM *);
  for(i = 0;i < nModes;i++) modeName[i] = mainDisasmTable[i]->name;
  retval = SelBoxA(modeName,nModes," Select disassembler: ",DefDisasmSel);
  if(retval != -1)
  {
    if(activeDisasm->term) activeDisasm->term();
    activeDisasm = mainDisasmTable[retval];
    DefDisasmSel = retval;
    disAcceptActions();
    return True;
  }
  return False;
}

static void __FASTCALL__ FillPrevAsmPage(unsigned long bound,unsigned predist)
{
  unsigned long distin,addr;
  unsigned j;
  unsigned totallen;
  char showref,addrdet;
  if(!predist) predist = Height*disMaxCodeLen;
  predist = (predist/16)*16+16;   /* align on 16-byte boundary */
  distin = bound >= predist ? bound - predist : 0;
  memset(PrevStrLenAddr,TW_DEF_FILLER,tvioHeight*sizeof(long));
  PrevStrCount = 0;
  totallen = 0;
  showref = disNeedRef;
  addrdet = hexAddressResolv;
  hexAddressResolv = 0;
  disNeedRef = 0;
  DisasmPrepareMode = True;
  for(j = 0;;j++)
  {
    DisasmRet dret;
       addr = distin + totallen;
       BMReadBufferEx(disCodeBuffer,disMaxCodeLen,addr,BM_SEEK_SET);
       dret = Disassembler(distin,(void *)disCodeBuffer,__DISF_SIZEONLY);
       if(addr >= bound) break;
       totallen += dret.codelen;
       if(j < Height) PrevStrLenAddr[j] = addr;
       else
       {
          memmove(PrevStrLenAddr,&PrevStrLenAddr[1],(Height - 1)*sizeof(long));
          PrevStrLenAddr[Height - 1] = addr;
       }
  }
  PrevStrCount = j < Height ? j : Height;
  if(PrevStrCount) LastPrevLen = bound - PrevStrLenAddr[PrevStrCount - 1];
  else             LastPrevLen = 0;
  disNeedRef = showref;
  DisasmPrepareMode = False;
  hexAddressResolv = addrdet;
}

static void __FASTCALL__ PrepareAsmLines(int keycode,unsigned long cfpos)
{
  switch(keycode)
  {
    case KE_DOWNARROW:
                    PrevStrLen = CurrStrLenBuff[0];
                    CurrStrLen = CurrStrLenBuff[1];
                    if(PrevStrCount < Height) PrevStrCount++;
                    else                      memmove(&PrevStrLenAddr[0],&PrevStrLenAddr[1],sizeof(long)*(Height - 1));
                    PrevStrLenAddr[PrevStrCount - 1] = cfpos - CurrStrLenBuff[0];
                    PrevPageSize = PrevStrLenAddr[PrevStrCount - 1] - PrevStrLenAddr[0] + PrevStrLen;
                    break;
    case KE_PGDN:
                    {
                      int i;
                      unsigned size;
                      unsigned long prevpos;
                      size = 0;
                      for(i = 0;i < Height;i++)
                      {
                         size += CurrStrLenBuff[i];
                      }
                      prevpos = cfpos - size;
                      size = 0;
                      for(i = 0;i < Height;i++)
                      {
                          size += CurrStrLenBuff[i];
                          PrevStrLenAddr[i] = prevpos + size;
                      }
                      PrevStrLen = CurrStrLenBuff[Height - 1];
                      PrevPageSize = size;
                      PrevStrCount = Height;
                    }
                    break;
   case KE_UPARROW:
                    FillPrevAsmPage(cfpos,(unsigned)(PrevPageSize + 15));
                    goto Calc;
    default:
                    FillPrevAsmPage(cfpos,0);
    Calc:
                    PrevPageSize = 0;
                    if(PrevStrCount)
                    {
                      PrevStrLen = LastPrevLen;
                      PrevPageSize = PrevStrLenAddr[PrevStrCount - 1] + LastPrevLen - PrevStrLenAddr[0];
                    }
                    else             PrevStrLen = 0;
                    break;
  }
}

static unsigned __FASTCALL__ drawAsm( unsigned keycode, unsigned textshift )
{
 int i,j,len,I,Limit,dir;
 unsigned long cfpos,flen,TopCFPos;
 ColorAttr insn_color;
 static unsigned long amocpos = 0L;
 char outstr[__TVIO_MAXSCREENWIDTH];
 HLInfo hli;
 cfpos = TopCFPos = BMGetCurrFilePos();
 if(keycode == KE_UPARROW)
 {
   char showref,addrdet;
   DisasmRet dret;
   showref = disNeedRef;
   addrdet = hexAddressResolv;
   disNeedRef = hexAddressResolv = 0;
   BMReadBufferEx(disCodeBuffer,disMaxCodeLen,cfpos,BM_SEEK_SET);
   DisasmPrepareMode = True;
   dret = Disassembler(cfpos,(void *)disCodeBuffer,__DISF_SIZEONLY);
   if(cfpos + dret.codelen != amocpos) keycode = KE_SUPERKEY;
   DisasmPrepareMode = False;
   disNeedRef = showref;
   hexAddressResolv = addrdet;
 }
 PrepareAsmLines(keycode,cfpos);
 if(keycode == KE_UPARROW) { I = Height - 1; dir = -1; Limit = -1; }
 else                   { I = 0; dir = 1; Limit = Height; }
 if(amocpos != cfpos || keycode == KE_SUPERKEY)
 {
   GidResetGoAddress(keycode);
   amocpos = cfpos;
   flen = BMGetFLength();
   twUseWin(MainWnd);
   for(i = I;i != Limit;i+=1*dir)
   {
     DisasmRet dret;
     memset(outstr,TW_DEF_FILLER,__TVIO_MAXSCREENWIDTH);
     DisasmCurrLine = i;
     if(cfpos < BMGetFLength())
     {
       if(keycode == KE_DOWNARROW && i < Height - 1)
       {
         tvioBuff it;
         t_vchar chars[__TVIO_MAXSCREENWIDTH];
         ColorAttr attrs[__TVIO_MAXSCREENWIDTH];
         it.chars = chars;
         it.attrs = attrs;
         CurrStrLenBuff[i] = CurrStrLenBuff[i + 1];
         twReadBuffer(MainWnd,1,i + 2,&it,Width);
         twWriteBuffer(MainWnd,1,i + 1,&it,Width);
         cfpos += CurrStrLenBuff[i];
       }
       else
       {
         if(!(keycode == KE_UPARROW && i > 0))
         {
           len = cfpos + disMaxCodeLen < flen ? disMaxCodeLen : (int)(flen - cfpos);
           memset(disCodeBuffer,0,disMaxCodeLen);
           BMReadBufferEx((void *)disCodeBuffer,len,cfpos,BM_SEEK_SET);
           dret = Disassembler(cfpos,(void *)disCodeBuffer,__DISF_NORMAL);
           if(i == 0) CurrStrLen = dret.codelen;
           CurrStrLenBuff[i] = dret.codelen;
           twSetColorAttr(browser_cset.main);
           memcpy(outstr,GidEncodeAddress(cfpos,hexAddressResolv),10);
           len = 0;
           if(disPanelMode < PANMOD_FULL)
           {
             static char _clone;
             twDirectWrite(1,i + 1,outstr,10);
             len = 10;
             twSetColorAttr(disasm_cset.family_id);
             _clone = activeDisasm->CloneShortName(dret.pro_clone);
             twDirectWrite(10,i + 1,&_clone,1);
             twSetColorAttr(browser_cset.main);
           }
           if(disPanelMode < PANMOD_MEDIUM)
           {
             int full_off,med_off;
             full_off = disMaxCodeLen*2+11;
             med_off = disMaxCodeLen*2+1;
             for(j = 0;j < dret.codelen;j++) { memcpy(&outstr[len],Get2Digit(disCodeBuffer[j]),2); len += 2; }
             if(len < (disPanelMode < PANMOD_FULL) ? full_off : med_off) len = disPanelMode < 2 ? full_off : med_off;
             twSetColorAttr(disasm_cset.opcodes);
             twDirectWrite(disPanelMode < PANMOD_FULL ? 11 : 1,
                           i + 1,
                           &outstr[10],
                           disPanelMode < PANMOD_FULL ? len - 11 : len - 1);
             if(isHOnLine(cfpos,dret.codelen))  
             {
                hli.text = &outstr[10];
                HiLightSearch(MainWnd,cfpos,10,dret.codelen,i,&hli,HLS_USE_DOUBLE_WIDTH);
             }
           }
           twSetColorAttr(browser_cset.main);
           twDirectWrite(len,i + 1," ",1);
           len++;
           insn_color = activeDisasm->GetInsnColor(dret.pro_clone);
           if(HiLight)
             twSetColorAttr(insn_color);
           else
             twSetColorAttr(browser_cset.main);
           j = strlen(dret.str);
           twDirectWrite(len,i + 1,dret.str,j); len += j;
           twSetColorAttr(browser_cset.main);
           if(len < Width)
           {
             twGotoXY(len,i + 1);
             twClrEOL();
           }
           cfpos += dret.codelen;
           BMSeek(cfpos,BM_SEEK_SET);
         }
         else
         {
           tvioBuff it;
           t_vchar chars[__TVIO_MAXSCREENWIDTH];
           ColorAttr attrs[__TVIO_MAXSCREENWIDTH];
           it.chars = chars;
           it.attrs = attrs;
           CurrStrLenBuff[i] = CurrStrLenBuff[i - 1];
           twReadBuffer(MainWnd,1,i,&it,Width);
           twWriteBuffer(MainWnd,1,i + 1,&it,Width);
         }
       }
     }
     else twDirectWrite(1,i + 1,outstr,Width);
   }
   twSetColorAttr(browser_cset.main);
   lastbyte = TopCFPos;
   for(i = I;i != Limit;i+=1*dir) lastbyte += CurrStrLenBuff[i];
   CurrPageSize = lastbyte-TopCFPos;
 }
 return textshift;
}

static unsigned long __FASTCALL__ disPrevPageSize( void ) { return PrevPageSize; }
static unsigned long __FASTCALL__ disCurrPageSize( void ) { return CurrPageSize; }
static unsigned long __FASTCALL__ disPrevLineWidth( void ) { return PrevStrLen; }
static unsigned long __FASTCALL__ disCurrLineWidth( void ) { return CurrStrLen; }
static const char *  __FASTCALL__ disMiscKeyName( void ) { return "Modify"; }

static const char *refsdepth_names[] =
{
   "~None",
   "~Calls/jmps only (navigation)",
   "~All"
};

static tBool __FASTCALL__ disReferenceResolving( void )
{
  size_t nModes;
  int retval;
  tBool ret;
  nModes = sizeof(refsdepth_names)/sizeof(char *);
  retval = SelBoxA(refsdepth_names,nModes," References resolving depth: ",disNeedRef);
  if(retval != -1)
  {
    disNeedRef = retval;
    ret = True;
  }
  else ret = False;
  if(detectedFormat->set_state)
    detectedFormat->set_state(disNeedRef ? PS_ACTIVE : PS_INACTIVE);
  return ret;
}

static const char *panmod_names[] =
{
   "~Full",
   "~Medium",
   "~Wide"
};

static tBool __FASTCALL__ disSelectPanelMode( void )
{
  unsigned nModes;
  int i;
  nModes = sizeof(panmod_names)/sizeof(char *);
  i = SelBoxA(panmod_names,nModes," Select panel mode: ",disPanelMode);
  if(i != -1)
  {
    disPanelMode = i;
    return True;
  }
  return False;
}

static const char *hilight_names[] =
{
   "~Mono",
   "~Highlight"
};

static tBool __FASTCALL__ disSelectHiLight( void )
{
  unsigned nModes;
  int i;
  nModes = sizeof(hilight_names)/sizeof(char *);
  i = SelBoxA(hilight_names,nModes," Select highlight mode: ",HiLight);
  if(i != -1)
  {
    HiLight = i;
    return True;
  }
  return False;
}

static tBool __FASTCALL__ disDetect( void ) { return True; }

static tBool __FASTCALL__ DefAsmAction(int _lastbyte,int start)
{
 int _index;
 tBool redraw,dox;
 char xx;
  redraw = True;
  xx = edit_x / 2;
  dox = True;
  _index = start + xx;
   switch(_lastbyte)
   {
     case KE_F(4)   : EditorMem.buff[_index] = ~EditorMem.buff[_index]; break;
     case KE_F(5)   : EditorMem.buff[_index] |= edit_XX; break;
     case KE_F(6)   : EditorMem.buff[_index] &= edit_XX; break;
     case KE_F(7)   : EditorMem.buff[_index] ^= edit_XX; break;
     case KE_F(8)   : EditorMem.buff[_index]  = edit_XX; break;
     case KE_F(9)   : EditorMem.buff[_index] = EditorMem.save[_index]; break;
     default        : redraw = edit_defaction(_lastbyte); dox = False; break;
   }
  if(dox) { xx++; edit_x = xx*2; }
  return redraw;
}

static void __FASTCALL__ DisasmScreen(TWindow* ewnd,unsigned long cp,unsigned long flen,int st,int stop,int start)
{
 int i,j,len,lim;
 char outstr[__TVIO_MAXSCREENWIDTH+1],outstr1[__TVIO_MAXSCREENWIDTH+1];
 DisasmRet dret;
 twFreezeWin(ewnd);
 for(i = st;i < stop;i++)
 {
  if(start + cp < flen)
  {
   memcpy(outstr,GidEncodeAddress(cp + start,hexAddressResolv),10);
   twUseWin(MainWnd);
   twSetColorAttr(browser_cset.main);
   twDirectWrite(1,i + 1,outstr,9);
   dret = Disassembler(cp + start,&EditorMem.buff[start],__DISF_NORMAL);
   EditorMem.alen[i] = dret.codelen;
   memset(outstr,TW_DEF_FILLER,Width);
   memset(outstr1,TW_DEF_FILLER,Width);
   len = 0; for(j = 0;j < EditorMem.alen[i];j++) { memcpy(&outstr1[len],Get2Digit(EditorMem.save[start + j]),2); len += 2; }
   len = 0; for(j = 0;j < EditorMem.alen[i];j++) { memcpy(&outstr[len],Get2Digit(EditorMem.buff[start + j]),2); len += 2; }
   twUseWin(ewnd);
   len = disMaxCodeLen*2;
   for(j = 0;j < len;j++)
   {
     if(outstr[j] == outstr1[j])
          twSetColorAttr(browser_cset.edit.main);
     else
          twSetColorAttr(browser_cset.edit.change);
     twDirectWrite(j + 1,i + 1,&outstr[j],1);
   }
   len = strlen(dret.str);
   memset(outstr,TW_DEF_FILLER,Width);
   memcpy(outstr,dret.str,len);
   twUseWin(MainWnd);
   twSetColorAttr(browser_cset.main);
   lim = disMaxCodeLen*2+11;
   twDirectWrite(lim+1,i + 1,outstr,Width-lim);
   start += EditorMem.alen[i];
  }
  else
  {
    twUseWin(MainWnd);
    twGotoXY(1,i + 1);
    twClrEOL();
    twUseWin(ewnd);
    twGotoXY(1,i + 1);
    twClrEOL();
    EditorMem.alen[i] = 0;
  }
 }
 twRefreshWin(ewnd);
}

static int __NEAR__ __FASTCALL__ FullAsmEdit(TWindow * ewnd)
{
 int j,_lastbyte,start;
 unsigned rlen,len,flags;
 unsigned long flen;
 unsigned max_buff_size = disMaxCodeLen*tvioHeight;
 tBool redraw = False;
 char outstr[__TVIO_MAXSCREENWIDTH],owork[__TVIO_MAXSCREENWIDTH];

 flen = BMGetFLength();
 edit_cp = BMGetCurrFilePos();
 start = 0;

 rlen = edit_cp + max_buff_size < flen ? max_buff_size : (int)(flen - edit_cp);
 BMReadBufferEx((void *)EditorMem.buff,rlen,edit_cp,BM_SEEK_SET);
 memcpy(EditorMem.save,EditorMem.buff,max_buff_size);
 memset(EditorMem.alen,TW_DEF_FILLER,tvioHeight);

 DisasmScreen(ewnd,edit_cp,flen,0,Height,start);
 PaintETitle(0,True);
 start = 0;
 twShowWin(ewnd);
 twSetCursorType(TW_CUR_NORM);
 while(1)
 {
   twUseWin(ewnd);

   len = 0; for(j = 0;j < EditorMem.alen[edit_y];j++) { memcpy(&owork[len],Get2Digit(EditorMem.save[start + j]),2); len += 2; }
   len = 0; for(j = 0;j < EditorMem.alen[edit_y];j++) { memcpy(&outstr[len],Get2Digit(EditorMem.buff[start + j]),2); len += 2; }
   flags = __ESS_FILLER_7BIT | __ESS_WANTRETURN | __ESS_HARDEDIT;
   if(!redraw) flags |= __ESS_NOREDRAW;
   _lastbyte = eeditstring(outstr,&legalchars[2],&len,(unsigned)(edit_y + 1),(unsigned *)&edit_x,
                          flags,owork);
   CompressHex(&EditorMem.buff[start],outstr,EditorMem.alen[edit_y],False);
   switch(_lastbyte)
   {
     case KE_TAB     :
                      {
                       AsmRet aret;
                       if(activeDisasm->asm_f)
                       {
                         char code[81];
                         if(GetStringDlg(code,activeDisasm->name,
                                         NULL,"Input assembler instruction:"))
                         {
                           aret = (*activeDisasm->asm_f)(code);
                           if(aret.err_code)
                           {
                              ErrMessageBox("Syntax error",NULL);
                              continue;
                           }
                           else
                           {
                              memcpy(outstr,aret.insn,aret.insn_len);
                           }
                         }
                         break;
                       }
                       else continue;
                      }
     case KE_F(1)    : ExtHelp(); continue;
     case KE_CTL_F(1): activeDisasm->action[0](); continue;
     case KE_CTL_F(2): SelectSysInfo(); continue;
     case KE_CTL_F(3): SelectTool(); continue;
     case KE_F(2)    :
                      {
                         BGLOBAL bHandle;
                         char *fname;
                         fname = BMName();
                         if((bHandle = biewOpenRW(fname,BBIO_SMALL_CACHE_SIZE)) != &bNull)
                         {
                           bioSeek(bHandle,edit_cp,BIO_SEEK_SET);
                           if(!bioWriteBuffer(bHandle,(void *)EditorMem.buff,rlen))
                              errnoMessageBox("Can not write",NULL,errno);
                           bioClose(bHandle);
                           BMReRead();
                         }
                         else errnoMessageBox("Can not reopen",NULL,errno);
                      }
     case KE_F(10)   :
     case KE_ESCAPE: goto bye;
     default: redraw = DefAsmAction(_lastbyte,start); break;
   }
   CheckYBounds();
   start = edit_y ? Summ(EditorMem.alen,edit_y) : 0;
   if(redraw)
   {
     DisasmRet dret;
     dret = Disassembler(edit_cp + start,&EditorMem.buff[start],__DISF_NORMAL);
     EditorMem.alen[edit_y] = dret.codelen;
     DisasmScreen(ewnd,edit_cp,flen,0,Height,0);
   }
   PaintETitle(start + edit_x/2,True);
   CheckXYBounds();
   start = edit_y ? Summ(EditorMem.alen,edit_y) : 0;
 }
 bye:
 twSetCursorType(TW_CUR_OFF);
 return _lastbyte;
}

static void __FASTCALL__ disEdit( void )
{
 TWindow * ewnd;
 ewnd = WindowOpen(11,2,disMaxCodeLen*2+11,tvioHeight-1,TWS_CURSORABLE);
 twSetColorAttr(browser_cset.edit.main); twClearWin();
 edit_x = edit_y = 0;
 EditMode = EditMode ? False : True;
 if(editInitBuffs(disMaxCodeLen))
 {
   FullAsmEdit(ewnd);
   editDestroyBuffs();
 }
 EditMode = EditMode ? False : True;
 CloseWnd(ewnd);
 PaintTitle();
}

static void __FASTCALL__ HelpAsm( void )
{
   if( activeDisasm->ShowShortHelp ) activeDisasm->ShowShortHelp();
}

static void __FASTCALL__ disReadIni( hIniProfile *ini )
{
  char tmps[10];
  if(isValidIniArgs())
  {
    biewReadProfileString(ini,"Biew","Browser","LastSubMode","0",tmps,sizeof(tmps));
    DefDisasmSel = (int)strtoul(tmps,NULL,10);
    if(DefDisasmSel >= sizeof(mainDisasmTable)/sizeof(REGISTRY_DISASM *)) DefDisasmSel = 0;
    ReadIniAResolv(ini);
    biewReadProfileString(ini,"Biew","Browser","SubSubMode7","0",tmps,sizeof(tmps));
    disPanelMode = (int)strtoul(tmps,NULL,10);
    if(disPanelMode > PANMOD_FULL) disPanelMode = 0;
    biewReadProfileString(ini,"Biew","Browser","SubSubMode8","0",tmps,sizeof(tmps));
    disNeedRef = (int)strtoul(tmps,NULL,10);
    if(disNeedRef > NEEDREF_ALL) disNeedRef = 0;
    biewReadProfileString(ini,"Biew","Browser","SubSubMode9","0",tmps,sizeof(tmps));
    HiLight = (int)strtoul(tmps,NULL,10);
    if(HiLight > 1) HiLight = 1;
    activeDisasm = mainDisasmTable[DefDisasmSel];
    disAcceptActions();
    if(activeDisasm->read_ini) activeDisasm->read_ini(ini);
  }
}

static void __FASTCALL__ disInit( void )
{
  int def_platform;
  CurrStrLenBuff = PMalloc(tvioHeight);
  PrevStrLenAddr = PMalloc(tvioHeight*sizeof(long));
  if((!CurrStrLenBuff) || (!PrevStrLenAddr))
  {
    err:
    MemOutBox("Disassembler initialization");
    exit(EXIT_FAILURE);
  }
  def_platform = DISASM_DEFAULT;
  if(detectedFormat->query_platform) def_platform = detectedFormat->query_platform();
  if(def_platform != DISASM_DEFAULT) DefDisasmSel = def_platform;
  activeDisasm = mainDisasmTable[DefDisasmSel];
  disAcceptActions();
  if(!initCodeGuider()) goto err;
}

static void __FASTCALL__ disTerm( void )
{
  termCodeGuider();
  if(activeDisasm->term) activeDisasm->term();
  PFREE(CurrStrLenBuff);
  PFREE(PrevStrLenAddr);
  PFREE(disCodeBuffer);
}

static void __FASTCALL__ disSaveIni( hIniProfile *ini )
{
  char tmps[10];
  sprintf(tmps,"%i",DefDisasmSel);
  biewWriteProfileString(ini,"Biew","Browser","LastSubMode",tmps);
  WriteIniAResolv(ini);
  sprintf(tmps,"%i",disPanelMode);
  biewWriteProfileString(ini,"Biew","Browser","SubSubMode7",tmps);
  sprintf(tmps,"%i",disNeedRef);
  biewWriteProfileString(ini,"Biew","Browser","SubSubMode8",tmps);
  sprintf(tmps,"%i",HiLight);
  biewWriteProfileString(ini,"Biew","Browser","SubSubMode9",tmps);
  if(activeDisasm->save_ini) activeDisasm->save_ini(ini);
}

DisasmRet Disassembler(unsigned long ulShift,MBuffer buffer,unsigned flags)
{
  return activeDisasm->disasm(ulShift,buffer,flags);
}

static unsigned __FASTCALL__ disCharSize( void ) { return 1; }

REGISTRY_MODE disMode =
{
  "~Dissasembler",
  { NULL, "Disasm", NULL, NULL, NULL, "AResol", "PanMod", "ResRef", "HiLght", NULL },
  { NULL, disSelect_Disasm, NULL, NULL, NULL, hexAddressResolution, disSelectPanelMode, disReferenceResolving, disSelectHiLight, NULL },
  disDetect,
  __MF_USECODEGUIDE,
  drawAsm,
  NULL,
  disCharSize,
  disMiscKeyName,
  disEdit,
  disPrevPageSize,
  disCurrPageSize,
  disPrevLineWidth,
  disCurrLineWidth,
  HelpAsm,
  disReadIni,
  disSaveIni,
  disInit,
  disTerm
};

static void __NEAR__ __FASTCALL__ disAcceptActions( void )
{
  if(activeDisasm->init) activeDisasm->init();
  disMaxCodeLen = activeDisasm->max_insn_len();
  disCodeBuffer = PMalloc(disMaxCodeLen);
  if(!disCodeBuffer)
  {
    MemOutBox("Disassembler initialization");
    exit(EXIT_FAILURE);
  }
  disMode.prompt[0] = activeDisasm->prompt[0];
  disMode.action[0] = activeDisasm->action[0];
  disMode.prompt[2] = activeDisasm->prompt[1];
  disMode.action[2] = activeDisasm->action[1];
  disMode.prompt[3] = activeDisasm->prompt[2];
  disMode.action[3] = activeDisasm->action[2];
  disMode.prompt[4] = activeDisasm->prompt[3];
  disMode.action[4] = activeDisasm->action[3];
  disMode.prompt[9] = activeDisasm->prompt[4];
  disMode.action[9] = activeDisasm->action[4];
}

/* Common disassembler utility */

char * __FASTCALL__ TabSpace(char * str,unsigned nSpace)
{
  int i,mx;
  unsigned len;
  len = strlen(str);
  mx = max(len,nSpace);
  for(i = len;i < mx;i++) str[i] = TW_DEF_FILLER;
  if(len > nSpace) str[i++] = TW_DEF_FILLER;
  str[i] = 0;
  return str;
}

void  __FASTCALL__ disSetModifier(char *str,const char *modf)
{
  unsigned i,len,mlen;
  i = 0;
  len = strlen(str);
  mlen = strlen(modf);
  while(str[i] != TW_DEF_FILLER) i++;
  i++;
  memcpy(&str[i],modf,mlen);
  if(i+mlen > len) { str[i+mlen] = TW_DEF_FILLER; str[i+mlen+1] = 0; }
}

void __FASTCALL__ disAppendDigits(char *str,unsigned long ulShift,int flags,
                     char codelen,void *defval,int type)
{
 tBool app;
 char *appstr;
#ifndef NDEBUG
  if(ulShift >= BMGetFLength()-codelen)
  {
     char sout[75];
     static tBool displayed = False;
     if(!displayed)
     {
       strncpy(sout,str,sizeof(sout)-1);
       sout[sizeof(sout)-1] = 0;
       ErrMessageBox(sout," Detected internal disassembler error ");
       displayed = True;
     }
  }
#endif
  if(disNeedRef == NEEDREF_ALL)
    app = AppendAsmRef(str,ulShift,flags,codelen,0L);
  else
    app = False;
  if(!app)
  {
    switch(type)
    {
      case DISARG_LONG:  appstr = Get8SignDig(*(long *)defval); break;
      case DISARG_SHORT: appstr = Get4SignDig(*(short *)defval); break;
      case DISARG_CHAR:  appstr = Get2SignDig(*(char *)defval); break;
      case DISARG_BYTE:  appstr = Get2Digit(*(unsigned char *)defval); break;
      case DISARG_WORD:  appstr = Get4Digit(*(unsigned short *)defval); break;
      default:
      case DISARG_DWORD: appstr = Get8Digit(*(unsigned long *)defval); break;
    }
    strcat(str,appstr);
  }
}

void __FASTCALL__ disAppendFAddr(char * str,long ulShift,long distin,unsigned long r_sh,char type,unsigned seg,char codelen)
{
 int needref;
 tBool appended = False;
#ifndef NDEBUG
  if(ulShift >= BMGetFLength()-codelen)
  {
     char sout[75];
     static tBool displayed = False;
     if(!displayed)
     {
       strncpy(sout,str,sizeof(sout)-1);
       sout[sizeof(sout)-1] = 0;
       ErrMessageBox(sout," Detected internal disassembler error ");
       displayed = True;
     }
  }
#endif
 if(disNeedRef > NEEDREF_NONE)
 {
   if(AppendAsmRef(str,ulShift,APREF_TRY_LABEL,codelen,r_sh)) appended = True;
 }
 if(!appended)
 {
   needref = disNeedRef;
   disNeedRef = NEEDREF_NONE;
   if(r_sh <= BMGetFLength())
   {
     const char * cptr;
     char lbuf[10];
     if(DumpMode) cptr = "L";
     else
     {
       cptr = "file:";
     }
     strcat(str,cptr);
     sprintf(lbuf,"%08lX",r_sh);
     strcat(str,lbuf);
     if(!DumpMode && !EditMode) GidAddGoAddress(str,r_sh);
   }
   else
   {
     const char * pstr = "";
     if(type & 0x04)
     {
       strcat(str,Get4Digit(seg));
       strcat(str,":");
     }
     if(!type) pstr = Get2SignDig((char)distin);
     else
       if(type & 0x01) pstr = type & 0x04 ? Get4Digit((unsigned)distin) : Get4SignDig((unsigned)distin);
       else
        if(type & 0x02)   pstr = Get8SignDig(distin);
     strcat(str,pstr);
   }
   disNeedRef = needref;
 }
}










